class QualityTask:
    """Quality class to be inherited by training_tasks or
    evaluation_tasks."""

    def __init__(self, json, client):
        self._json = json
        self.id = json["id"]
        self.initial_response = getattr(json, "initial_response", None)
        self.expected_response = json["expected_response"]
        self._client = client

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"QualityTask(id={self.id})"

    def __repr__(self):
        return f"QualityTask({self._json})"

    def as_dict(self):
        """Returns all attributes as a dictionary"""
        return self._json



from scaleapi.quality_tasks import QualityTask


class TrainingTask(QualityTask):
    """TrainingTask class, containing TrainingTask information."""

    def __str__(self):
        return f"TrainingTask(id={self.id})"

    def __repr__(self):
        return f"TrainingTask({self._json})"



from enum import Enum


class BatchStatus(Enum):
    """Status of Batches"""

    Staging = "staging"
    InProgress = "in_progress"
    Completed = "completed"


class Batch:
    """Batch class, contains Batch information"""

    def __init__(self, json, client):
        self._json = json
        self.name = json["name"]
        self.status = json["status"]
        self.project = json["project"]
        self.created_at = json["created_at"]
        self.project = json["project"]

        self.tasks_pending = None
        self.tasks_completed = None
        self.tasks_error = None
        self.tasks_canceled = None
        self._client = client

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return f"Batch(name={self.name})"

    def __repr__(self):
        return f"Batch({self._json})"

    def as_dict(self):
        """Returns all attributes as a dictionary"""
        return self._json

    def finalize(self):
        """Finalizes the batch"""
        res = self._client.finalize_batch(self.name)
        self.status = res.status
        return res

    def get_status(self):
        """Returns status of the batch and
        updates tasks_... parameters
        """
        res = self._client.batch_status(self.name)
        self.status = res["status"]
        for stat in ["pending", "completed", "error", "canceled"]:
            setattr(self, "tasks_" + stat, res.get(stat, 0))
        return res



from typing import Dict


class ScaleException(Exception):
    """Generic ScaleException class"""

    code = None

    def __init__(self, message, errcode=None):
        if not message:
            message = type(self).__name__
        self.message = message

        if errcode:
            self.code = errcode

        if self.code:
            super().__init__(f"<Response [{self.code}]> {message}")
        else:
            super().__init__(f"<Response> {message}")


class ScaleInvalidRequest(ScaleException):
    """400 - Bad Request -- The request was unacceptable,
    often due to missing a required parameter.
    """

    code = 400


class ScaleUnauthorized(ScaleException):
    """401 - Unauthorized -- No valid API key provided."""

    code = 401


class ScaleNotEnabled(ScaleException):
    """402 - Not enabled -- Please contact sales@scaleapi.com before
    creating this type of task.
    """

    code = 402


class ScaleResourceNotFound(ScaleException):
    """404 - Not Found -- The requested resource doesn't exist."""

    code = 404


class ScaleDuplicateResource(ScaleException):
    """409 - Conflict -- Object already exists with same name,
    idempotency key or unique_id.
    """

    code = 409


class ScaleTooManyRequests(ScaleException):
    """429 - Too Many Requests -- Too many requests hit the API
    too quickly.
    """

    code = 429


class ScaleInternalError(ScaleException):
    """500 - Internal Server Error -- We had a problem with our server.
    Try again later.
    """

    code = 500


class ScaleServiceUnavailable(ScaleException):
    """503 - Server Timeout From Request Queueing -- Try again later."""

    code = 503


class ScaleTimeoutError(ScaleException):
    """504 - Server Timeout Error -- Try again later."""

    code = 504


ExceptionMap: Dict[int, ScaleException] = {
    ScaleInvalidRequest.code: ScaleInvalidRequest,
    ScaleUnauthorized.code: ScaleUnauthorized,
    ScaleNotEnabled.code: ScaleNotEnabled,
    ScaleResourceNotFound.code: ScaleResourceNotFound,
    ScaleDuplicateResource.code: ScaleDuplicateResource,
    ScaleTooManyRequests.code: ScaleTooManyRequests,
    ScaleInternalError.code: ScaleInternalError,
    ScaleTimeoutError.code: ScaleTimeoutError,
    ScaleServiceUnavailable.code: ScaleServiceUnavailable,
}



class Project:
    """Project class, containing Project information."""

    def __init__(self, json, client):
        self._json = json
        self.name = json["name"]
        self.type = json["type"]
        self._client = client
        self.params = None
        self.version = None
        self.instruction = None

        if len(json["param_history"]):
            self.params = json["param_history"][-1]
            self.version = self.params["version"]
            if "instruction" in self.params:
                self.instruction = self.params["instruction"]

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return f"Project(name={self.name})"

    def __repr__(self):
        return f"Project({self._json})"

    def as_dict(self):
        """Returns all attributes as a dictionary"""
        return self._json



class File:
    """File class, containing File information."""

    def __init__(self, json, client):
        self._json = json
        self.id = json["id"]
        self.attachment_url = json["attachment_url"]
        self._client = client

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"File(id={self.id})"

    def __repr__(self):
        return f"File({self._json})"

    def as_dict(self):
        """Returns all attributes as a dictionary"""
        return self._json



from scaleapi.quality_tasks import QualityTask


class EvaluationTask(QualityTask):
    """EvaluationTask class, containing EvaluationTask information."""

    def __str__(self):
        return f"EvaluationTask(id={self.id})"

    def __repr__(self):
        return f"EvaluationTask({self._json})"



import platform
import urllib.parse

import requests
from requests.adapters import HTTPAdapter, Response, Retry

from ._version import __package_name__, __version__
from .exceptions import ExceptionMap, ScaleException

SCALE_API_BASE_URL_V1 = "https://api.scale.com/v1"

# Parameters for HTTP retry
HTTP_TOTAL_RETRIES = 3  # Number of total retries
HTTP_RETRY_BACKOFF_FACTOR = 2  # Wait 1, 2, 4 seconds between retries
HTTP_STATUS_FORCE_LIST = [408, 429] + list(range(500, 531))
HTTP_RETRY_ALLOWED_METHODS = frozenset({"GET", "POST", "DELETE"})


class Api:
    """Internal Api reference for handling http operations"""

    def __init__(self, api_key, user_agent_extension=None, api_instance_url=None):
        if api_key == "" or api_key is None:
            raise ScaleException("Please provide a valid API Key.")

        self.api_key = api_key

        self._auth = (self.api_key, "")
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": self._generate_useragent(user_agent_extension),
        }
        self._headers_multipart_form_data = {
            "User-Agent": self._generate_useragent(user_agent_extension),
        }
        self.base_api_url = api_instance_url or SCALE_API_BASE_URL_V1

    @staticmethod
    def _http_request(
        method,
        url,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ) -> Response:

        https = requests.Session()
        retry_strategy = Retry(
            total=HTTP_TOTAL_RETRIES,
            backoff_factor=HTTP_RETRY_BACKOFF_FACTOR,
            status_forcelist=HTTP_STATUS_FORCE_LIST,
            allowed_methods=HTTP_RETRY_ALLOWED_METHODS,
            raise_on_status=False,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        https.mount("https://", adapter)

        try:
            params = params or {}
            body = body or None

            res = https.request(
                method=method,
                url=url,
                headers=headers,
                auth=auth,
                params=params,
                json=body,
                files=files,
                data=data,
            )

            return res
        except Exception as err:
            raise ScaleException(err) from err

    @staticmethod
    def _raise_on_respose(res: Response):

        try:
            message = res.json().get("error", res.text)
        except ValueError:
            message = res.text

        exception = ExceptionMap.get(res.status_code, ScaleException)
        raise exception(message, res.status_code)

    def _api_request(
        self,
        method,
        endpoint,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ):
        """Generic HTTP request method with error handling."""

        url = f"{self.base_api_url}/{endpoint}"

        res = self._http_request(method, url, headers, auth, params, body, files, data)

        json = None
        if res.status_code == 200:
            json = res.json()
        else:
            self._raise_on_respose(res)

        return json

    def get_request(self, endpoint, params=None):
        """Generic GET Request Wrapper"""
        return self._api_request(
            "GET", endpoint, headers=self._headers, auth=self._auth, params=params
        )

    def post_request(self, endpoint, body=None, files=None, data=None):
        """Generic POST Request Wrapper"""
        return self._api_request(
            "POST",
            endpoint,
            headers=self._headers
            if files is None
            else self._headers_multipart_form_data,
            auth=self._auth,
            body=body,
            files=files,
            data=data,
        )

    def delete_request(self, endpoint, params=None):
        """Generic DELETE Request Wrapper"""
        return self._api_request(
            "DELETE", endpoint, headers=self._headers, auth=self._auth, params=params
        )

    @staticmethod
    def _generate_useragent(extension: str = None) -> str:
        """Generates UserAgent parameter with module, Python
        and OS details

        Args:
            extension (str, optional): Option to extend UserAgent
            with source system

        Returns:
            str: Generated UserAgent parameter with platform versions
        """
        python_version = platform.python_version()
        os_platform = platform.platform()

        user_agent = " ".join(
            filter(
                None,
                [
                    f"{__package_name__}/{__version__}",
                    f"Python/{python_version}",
                    f"OS/{os_platform}",
                    extension,
                ],
            )
        )
        return user_agent

    @staticmethod
    def quote_string(text: str) -> str:
        """Project and Batch names can be a part of URL, which causes
        an error in case of a special character used.
        Quotation assures the right object to be retrieved from API.

        `quote_string('a bc/def')` -> `a%20bc%2Fdef`

        Args:
            text (str): Input text to be quoted

        Returns:
            str: Quoted text in return
        """
        return urllib.parse.quote(text, safe="")



import os.path

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

install_requires = ["requests>=2.25.0", "urllib3>=1.26.0"]


def read(rel_path):
    """Read lines from given file"""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    """Read __version__ from given file"""
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError(f"Unable to find a valid __version__ string in {rel_path}.")


setup(
    name="scaleapi",
    packages=["scaleapi"],
    version=get_version("scaleapi/_version.py"),
    description="The official Python client library for Scale AI, "
    "the Data Platform for AI",
    author="Scale AI",
    author_email="support@scale.com",
    url="https://github.com/scaleapi/scaleapi-python-client",
    keywords=[
        "scale",
        "scaleapi",
        "tasks",
        "categorization",
        "labeling",
        "annotation",
    ],
    install_requires=install_requires,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
    ],
)



from typing import IO, Dict, Generator, Generic, List, TypeVar, Union

from scaleapi.batches import Batch, BatchStatus
from scaleapi.evaluation_tasks import EvaluationTask
from scaleapi.exceptions import ScaleInvalidRequest
from scaleapi.files import File
from scaleapi.projects import Project
from scaleapi.training_tasks import TrainingTask

from ._version import __version__  # noqa: F401
from .api import Api
from .tasks import Task, TaskReviewStatus, TaskStatus, TaskType

T = TypeVar("T")


class Paginator(list, Generic[T]):
    """Paginator for list endpoints"""

    def __init__(
        self,
        docs: List[T],
        total: int,
        limit: int,
        offset: int,
        has_more: bool,
        next_token=None,
    ):
        super().__init__(docs)
        self.docs = docs
        self.total = total
        self.limit = limit
        self.offset = offset
        self.has_more = has_more
        self.next_token = next_token


class Tasklist(Paginator[Task]):
    """Tasks Paginator"""


class Batchlist(Paginator[Batch]):
    """Batches Paginator"""


class ScaleClient:
    """Main class serves as an interface for Scale API"""

    def __init__(self, api_key, source=None, api_instance_url=None):
        self.api = Api(
            api_key, user_agent_extension=source, api_instance_url=api_instance_url
        )

    def get_task(self, task_id: str) -> Task:
        """Fetches a task.
        Returns the associated task.

        Args:
            task_id (str):
                Task identifier
        Returns:
            Task:
        """
        endpoint = f"task/{task_id}"
        return Task(self.api.get_request(endpoint), self)

    def cancel_task(self, task_id: str, clear_unique_id: bool = False) -> Task:
        """Cancels a task and returns the associated task.
        Raises a ScaleException if it has already been canceled.

        Args:
            task_id (str):
                Task id
            clear_unique_id (boolean):
                Option to clear unique id when the task is deleted

        Returns:
            Task
        """
        if clear_unique_id:
            endpoint = f"task/{task_id}/cancel?clear_unique_id=true"
        else:
            endpoint = f"task/{task_id}/cancel"
        return Task(self.api.post_request(endpoint), self)

    def update_task_unique_id(self, task_id: str, unique_id: str) -> Task:
        """Updates a task's unique_id and returns the associated task.
        Raises a ScaleDuplicateResource exception if unique_id
        is already in use.

        Args:
            task_id (str):
                Task id
            unique_id (str):
                unique_id to set

        Returns:
            Task
        """
        payload = dict(unique_id=unique_id)
        endpoint = f"task/{task_id}/unique_id"
        return Task(self.api.post_request(endpoint, body=payload), self)

    def clear_task_unique_id(self, task_id: str) -> Task:
        """Clears a task's unique_id and returns the associated task.

        Args:
            task_id (str):
                Task id

        Returns:
            Task
        """
        endpoint = f"task/{task_id}/unique_id"
        return Task(self.api.delete_request(endpoint), self)

    def tasks(self, **kwargs) -> Tasklist:
        """Returns a list of your tasks.
        Returns up to 100 at a time, to get more, use the
        next_token param passed back.

        Valid Args:
            start_time (str):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            end_time (str):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            status (str):
                Status to filter tasks, can be 'completed', 'pending',
                or 'canceled'

            type (str):
                Task type to filter. i.e. 'imageannotation'

            project (str):
                Project name to filter tasks by

            batch (str):
                Batch name to filter tasks by

            customer_review_status (str):
                Audit status of task, can be 'pending', 'fixed',
                'accepted' or 'rejected'.

            unique_id (List[str] | str):
                The unique_id of a task.

            completed_after (str):
                The minimum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            completed_before (str):
                The maximum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_after (str):
                The minimum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_before (str):
                The maximum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_after (str):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_before (str):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            tags (List[str] | str):
                The tags of a task; multiple tags can be
                specified as a list.

            limit (int):
                Determines the page size (1-100)

            next_token (str):
                Can be use to fetch the next page of tasks
        """
        allowed_kwargs = {
            "start_time",
            "end_time",
            "status",
            "type",
            "project",
            "batch",
            "limit",
            "completed_before",
            "completed_after",
            "next_token",
            "customer_review_status",
            "tags",
            "updated_before",
            "updated_after",
            "unique_id",
        }

        for key in kwargs:
            if key not in allowed_kwargs:
                raise ScaleInvalidRequest(
                    f"Illegal parameter {key} for ScaleClient.tasks()"
                )

        response = self.api.get_request("tasks", params=kwargs)

        docs = [Task(json, self) for json in response["docs"]]
        return Tasklist(
            docs,
            response["total"],
            response["limit"],
            response["offset"],
            response["has_more"],
            response.get("next_token"),
        )

    def get_tasks(
        self,
        project_name: str,
        batch_name: str = None,
        task_type: TaskType = None,
        status: TaskStatus = None,
        review_status: Union[List[TaskReviewStatus], TaskReviewStatus] = None,
        unique_id: Union[List[str], str] = None,
        completed_after: str = None,
        completed_before: str = None,
        updated_after: str = None,
        updated_before: str = None,
        created_after: str = None,
        created_before: str = None,
        tags: Union[List[str], str] = None,
    ) -> Generator[Task, None, None]:
        """Retrieve all tasks as a `generator` method, with the
        given parameters. This methods handles pagination of
        tasks() method.

        In order to retrieve results as a list, please use:
        `task_list = list(get_tasks(...))`

        Args:
            project_name (str):
                Project Name

            batch_name (str, optional):
                Batch Name

            task_type (TaskType, optional):
                Task type to filter i.e. `TaskType.TextCollection`

            status (TaskStatus, optional):
                Task status i.e. `TaskStatus.Completed`

            review_status (List[TaskReviewStatus] | TaskReviewStatus):
                The status of the audit result of the task.
                Input can be a single element or a list of
                TaskReviewStatus. i.e. `TaskReviewStatus.Accepted` to
                filter the tasks that you accepted after audit.

            unique_id (List[str] | str, optional):
                The unique_id of a task. Multiple unique IDs can be
                specified at the same time as a list.

            completed_after (str, optional):
                The minimum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            completed_before (str, optional):
                The maximum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_after (str, optional):
                The minimum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_before (str, optional):
                The maximum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_after (str, optional):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_before (str, optional):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            tags (List[str] | str, optional):
                The tags of a task; multiple tags can be
                specified as a list.

        Yields:
            Generator[Task]:
                Yields Task objects, can be iterated.
        """

        next_token = None
        has_more = True

        tasks_args = self._process_tasks_endpoint_args(
            project_name,
            batch_name,
            task_type,
            status,
            review_status,
            unique_id,
            completed_after,
            completed_before,
            updated_after,
            updated_before,
            created_after,
            created_before,
            tags,
        )

        while has_more:
            tasks_args["next_token"] = next_token

            tasks = self.tasks(**tasks_args)
            for task in tasks.docs:
                yield task

            next_token = tasks.next_token
            has_more = tasks.has_more

    def get_tasks_count(
        self,
        project_name: str,
        batch_name: str = None,
        task_type: TaskType = None,
        status: TaskStatus = None,
        review_status: Union[List[TaskReviewStatus], TaskReviewStatus] = None,
        unique_id: Union[List[str], str] = None,
        completed_after: str = None,
        completed_before: str = None,
        updated_after: str = None,
        updated_before: str = None,
        created_after: str = None,
        created_before: str = None,
        tags: Union[List[str], str] = None,
    ) -> int:
        """Returns number of tasks with given filters.

        Args:
            project_name (str):
                Project Name

            batch_name (str, optional):
                Batch Name

            task_type (TaskType, optional):
                Task type to filter i.e. `TaskType.TextCollection`

            status (TaskStatus, optional):
                Task status i.e. `TaskStatus.Completed`

            review_status (List[TaskReviewStatus] | TaskReviewStatus):
                The status of the audit result of the task.
                Input can be a single element or a list of
                TaskReviewStatus. i.e. `TaskReviewStatus.Accepted` to
                filter the tasks that you accepted after audit.

            unique_id (List[str] | str, optional):
                The unique_id of a task. Multiple unique IDs can be
                specified at the same time as a list.

            completed_after (str, optional):
                The minimum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            completed_before (str, optional):
                The maximum value of `completed_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_after (str, optional):
                The minimum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            updated_before (str, optional):
                The maximum value of `updated_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_after (str, optional):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_before (str, optional):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            tags (List[str] | str, optional):
                The tags of a task; multiple tags can be
                specified as a list.

        Returns:
            int:
                Returns number of tasks
        """

        tasks_args = self._process_tasks_endpoint_args(
            project_name,
            batch_name,
            task_type,
            status,
            review_status,
            unique_id,
            completed_after,
            completed_before,
            updated_after,
            updated_before,
            created_after,
            created_before,
            tags,
        )

        tasks_args["limit"] = 1

        tasks = self.tasks(**tasks_args)
        return tasks.total

    @staticmethod
    def _process_tasks_endpoint_args(
        project_name: str,
        batch_name: str = None,
        task_type: TaskType = None,
        status: TaskStatus = None,
        review_status: Union[List[TaskReviewStatus], TaskReviewStatus] = None,
        unique_id: Union[List[str], str] = None,
        completed_after: str = None,
        completed_before: str = None,
        updated_after: str = None,
        updated_before: str = None,
        created_after: str = None,
        created_before: str = None,
        tags: Union[List[str], str] = None,
    ):
        """Generates args for /tasks endpoint."""
        tasks_args = {
            "start_time": created_after,
            "end_time": created_before,
            "project": project_name,
            "batch": batch_name,
            "completed_before": completed_before,
            "completed_after": completed_after,
            "tags": tags,
            "updated_before": updated_before,
            "updated_after": updated_after,
            "unique_id": unique_id,
        }

        if status:
            tasks_args["status"] = status.value
        if task_type:
            tasks_args["type"] = task_type.value
        if review_status:
            if isinstance(review_status, List):
                value = ",".join(map(lambda x: x.value, review_status))
            else:
                value = review_status.value

            tasks_args["customer_review_status"] = value

        return tasks_args

    def create_task(self, task_type: TaskType, **kwargs) -> Task:
        """This method can be used for any Scale supported task type.
        Parameters may differ based on the given task_type.
        https://github.com/scaleapi/scaleapi-python-client#create-task

        Args:
            task_type (TaskType):
                Task type to be created
                i.e. `TaskType.ImageAnnotation`
            **kwargs:
                Passing in the applicable values into thefunction
                definition. The applicable fields and further
                information for each task type can be found in
                Scale's API documentation.
                https://docs.scale.com/reference

        Returns:
            Task:
                Returns created task.
        """
        endpoint = f"task/{task_type.value}"
        taskdata = self.api.post_request(endpoint, body=kwargs)
        return Task(taskdata, self)

    def create_batch(
        self,
        project: str,
        batch_name: str,
        callback: str = "",
        calibration_batch: bool = False,
        self_label_batch: bool = False,
    ) -> Batch:
        """Create a new Batch within a project.
        https://docs.scale.com/reference#batch-creation

        Args:
            project (str):
                Project name to create batch in
            batch_name (str):
                Batch name
            callback (str, optional):
                Email to notify, or URL to POST to
                when a batch is complete.
            calibration_batch (bool):
                Only applicable for self serve projects.
                Create a calibration_batch batch by setting
                the calibration_batch flag to true.
            self_label_batch (bool):
                Only applicable for self serve projects.
                Create a self_label batch by setting
                the self_label_batch flag to true.

        Returns:
            Batch: Created batch object
        """
        endpoint = "batches"
        payload = dict(
            project=project,
            name=batch_name,
            calibration_batch=calibration_batch,
            self_label_batch=self_label_batch,
            callback=callback,
        )
        batchdata = self.api.post_request(endpoint, body=payload)
        return Batch(batchdata, self)

    def finalize_batch(self, batch_name: str) -> Batch:
        """Finalizes a batch so its tasks can be worked on.
        https://docs.scale.com/reference#batch-finalization

        Args:
            batch_name (str):
                Batch name

        Returns:
            Batch
        """
        endpoint = f"batches/{Api.quote_string(batch_name)}/finalize"
        batchdata = self.api.post_request(endpoint)
        return Batch(batchdata, self)

    def batch_status(self, batch_name: str) -> Dict:
        """Returns the status of a batch with the counts of
        its tasks grouped by task status.
        https://docs.scale.com/reference#batch-status

        Args:
            batch_name (str):
                Batch name

        Returns:
            Dict {
                status: Batch status
                pending (optional): # of tasks in pending stage
                error (optional): # of tasks in error stage
                completed (optional): # of tasks in completed stage
                canceled (optional): # of tasks in canceled stage
            }
        """
        endpoint = f"batches/{Api.quote_string(batch_name)}/status"
        status_data = self.api.get_request(endpoint)
        return status_data

    def get_batch(self, batch_name: str) -> Batch:
        """Returns the details of a batch with the given name.
        https://docs.scale.com/reference#batch-retrieval

        Args:
            batch_name (str):
                Batch name

        Returns:
            Batch
        """
        endpoint = f"batches/{Api.quote_string(batch_name)}"
        batchdata = self.api.get_request(endpoint)
        return Batch(batchdata, self)

    def batches(self, **kwargs) -> Batchlist:
        """This is a paged endpoint for all of your batches.
        Pagination is based off limit and offset parameters,
        which determine the page size and how many results to skip.
        Returns up to 100 batches at a time (limit).
        https://docs.scale.com/reference#batch-list

        Valid Args:
            start_time (str):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            end_time (str):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            status (str):
                Status to filter batches by

            project (str):
                Project name to filter batches by

            limit (int):
                Determines the page size (1-100)

            offset (int):
                How many results to skip

        Returns:
            Batchlist:
                Paginated result. Batchlist.docs provides access
                to batches list. Batchlist.limit and Batchlist.offset
                are helpers for pagination.
        """
        allowed_kwargs = {
            "start_time",
            "end_time",
            "status",
            "project",
            "limit",
            "offset",
        }

        for key in kwargs:
            if key not in allowed_kwargs:
                raise ScaleInvalidRequest(
                    f"Illegal parameter {key} for ScaleClient.batches()"
                )
        endpoint = "batches"
        response = self.api.get_request(endpoint, params=kwargs)
        docs = [Batch(doc, self) for doc in response["docs"]]

        return Batchlist(
            docs,
            response["totalDocs"],
            response["limit"],
            response["offset"],
            response["has_more"],
        )

    def get_batches(
        self,
        project_name: str = None,
        batch_status: BatchStatus = None,
        created_after: str = None,
        created_before: str = None,
    ) -> Generator[Batch, None, None]:
        """`Generator` method to yield all batches with the given
        parameters.

        In order to retrieve results as a list, please use:
        `batches_list = list(get_batches(...))`

        Args:
            project_name (str):
                Project Name to filter batches

            batch_status (BatchStatus, optional):
                i.e. `BatchStatus.Completed`

            created_after (str, optional):
                The minimum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

            created_before (str, optional):
                The maximum value of `created_at` in UTC timezone
                ISO format: 'YYYY-MM-DD HH:MM:SS.mmmmmm'

        Yields:
            Generator[Batch]:
                Yields Batch, can be iterated.
        """

        has_more = True
        offset = 0

        while has_more:
            batches_args = {
                "start_time": created_after,
                "end_time": created_before,
                "project": project_name,
                "offset": offset,
            }

            if batch_status:
                batches_args["status"] = batch_status.value

            batches = self.batches(**batches_args)
            for batch in batches.docs:
                yield batch
            offset += batches.limit
            has_more = batches.has_more

    def create_project(
        self,
        project_name: str,
        task_type: TaskType,
        params: Dict = None,
        rapid: bool = False,
        studio: bool = False,
    ) -> Project:
        """Creates a new project.
        https://docs.scale.com/reference#project-creation

        Args:
            project_name (str):
                Project name

            task_type (TaskType):
                Task Type i.e. `TaskType.ImageAnnotation`

            params (Dict):
                Project parameters to be specificed.
                i.e. `{'instruction':'Please label the kittens'}`

        Returns:
            Project: [description]
        """
        endpoint = "projects"
        payload = dict(
            type=task_type.value,
            name=project_name,
            params=params,
            rapid=rapid,
            studio=studio,
        )
        projectdata = self.api.post_request(endpoint, body=payload)
        return Project(projectdata, self)

    def get_project(self, project_name: str) -> Project:
        """Retrieves a single project with the given name.
        https://docs.scale.com/reference#project-retrieval

        Args:
            project_name (str):
                Project name

        Returns:
            Project
        """
        endpoint = f"projects/{Api.quote_string(project_name)}"
        projectdata = self.api.get_request(endpoint)
        return Project(projectdata, self)

    def get_projects(self) -> List[Project]:
        """Returns all projects.
        Refer to Projects API Reference:
        https://docs.scale.com/reference#list-all-projects
        Same as `projects()` method.

        Returns:
            List[Project]
        """
        return self.projects()

    def projects(self) -> List[Project]:
        """Returns all projects.
        Refer to Projects API Reference:
        https://docs.scale.com/reference#list-all-projects

        Returns:
            List[Project]
        """
        endpoint = "projects"
        project_list = self.api.get_request(endpoint)
        return [Project(project, self) for project in project_list]

    def update_project(self, project_name: str, **kwargs) -> Project:
        """You can set parameters on a project. Project-level-parameters
        will be set on future tasks created under this project if they
        are not set in the task request. Any parameters specified in
        the task request will override any project parameter.
        https://docs.scale.com/reference#project-update-parameters

        Args:
            project_name (str):
                Project's name

            **kwargs:
                Project parameters to be set.

        Returns:
            Project
        """

        endpoint = f"projects/{Api.quote_string(project_name)}/setParams"
        projectdata = self.api.post_request(endpoint, body=kwargs)
        return Project(projectdata, self)

    def upload_file(self, file: IO, **kwargs) -> File:
        """Upload file.
        Refer to Files API Reference:
        https://docs.scale.com/reference#file-upload-1

        Args:
            file (IO):
                File buffer

        Returns:
            File
        """

        endpoint = "files/upload"
        files = {"file": file}
        filedata = self.api.post_request(endpoint, files=files, data=kwargs)
        return File(filedata, self)

    def import_file(self, file_url: str, **kwargs) -> File:
        """Import file from a remote url.
        Refer to Files API Reference:
        https://docs.scale.com/reference#file-import-1

        Args:
            file_url (str):
                File's url

        Returns:
            File
        """

        endpoint = "files/import"
        payload = dict(file_url=file_url, **kwargs)
        filedata = self.api.post_request(endpoint, body=payload)
        return File(filedata, self)

    def create_evaluation_task(
        self,
        task_type: TaskType,
        **kwargs,
    ) -> EvaluationTask:
        """This method can only be used for Rapid projects.
        Supported Task Types: [
            DocumentTranscription,
            SegmentAnnotation,
            VideoPlaybackAnnotation,
            ImageAnnotation,
            TextCollection,
            NamedEntityRecognition
        ]
        Parameters may differ based on the given task_type.

        Args:
            task_type (TaskType):
                Task type to be created
                e.g.. `TaskType.ImageAnnotation`
            **kwargs:
                The same set of parameters are expected with
                create_task function. Additionally with
                an expected_response and an optional initial_response
                if you want to make it a review phase evaluation task
                The expected_response/initial_response should follow
                the format of any other tasks' response on your project.
                It's recommended to try a self_label batch to get
                familiar with the response format.
                Scale's API documentation.
                https://docs.scale.com/reference

        Returns:
            EvaluationTask:
                Returns created evaluation task.
        """
        endpoint = f"evaluation_tasks/{task_type.value}"

        evaluation_task_data = self.api.post_request(endpoint, body=kwargs)
        return EvaluationTask(evaluation_task_data, self)

    def create_training_task(
        self,
        task_type: TaskType,
        **kwargs,
    ) -> TrainingTask:
        """This method can only be used for Rapid projects.
        Supported Task Types: [
            DocumentTranscription,
            SegmentAnnotation,
            VideoPlaybackAnnotation,
            ImageAnnotation,
            TextCollection,
            NamedEntityRecognition
        ]
        Parameters may differ based on the given task_type.

        Args:
            task_type (TaskType):
                Task type to be created
                e.g.. `TaskType.ImageAnnotation`
            **kwargs:
                The same set of parameters are expected with
                create_task function and an additional
                expected_response. Scale's API documentation.
                https://docs.scale.com/reference

        Returns:
            TrainingTask:
                Returns created training task.
        """
        endpoint = f"training_tasks/{task_type.value}"

        training_task_data = self.api.post_request(endpoint, body=kwargs)
        return TrainingTask(training_task_data, self)



from enum import Enum


class TaskType(Enum):
    """Task Type List"""

    Annotation = "annotation"
    Categorization = "categorization"
    Comparison = "comparison"
    CuboidAnnotation = "cuboidannotation"
    DataCollection = "datacollection"
    DocumentModel = "documentmodel"
    DocumentTranscription = "documenttranscription"
    ImageAnnotation = "imageannotation"
    LaneAnnotation = "laneannotation"
    LidarAnnotation = "lidarannotation"
    LidarLinking = "lidarlinking"
    LidarSegmentation = "lidarsegmentation"
    LidarTopdown = "lidartopdown"
    LineAnnotation = "lineannotation"
    NamedEntityRecognition = "namedentityrecognition"
    PointAnnotation = "pointannotation"
    PolygonAnnotation = "polygonannotation"
    SegmentAnnotation = "segmentannotation"
    Transcription = "transcription"
    TextCollection = "textcollection"
    VideoAnnotation = "videoannotation"
    VideoBoxAnnotation = "videoboxannotation"
    VideoPlaybackAnnotation = "videoplaybackannotation"
    VideoCuboidAnnotation = "videocuboidannotation"


class TaskReviewStatus(Enum):
    """Customer Audit Status of Task"""

    Accepted = "accepted"
    Fixed = "fixed"
    Commented = "commented"
    Rejected = "rejected"
    Pending = "pending"


class TaskStatus(Enum):
    """Status of Task"""

    Pending = "pending"
    Completed = "completed"
    Canceled = "canceled"


class Task:
    """Task class, containing task information."""

    def __init__(self, json, client):
        self._client = client
        self._json = json
        self.id = json["task_id"]

    def __getattr__(self, name):
        if name in self._json:
            return self._json[name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute {name}")

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"Task(id={self.id})"

    def __repr__(self):
        return f"Task({self._json})"

    def as_dict(self):
        """Returns object details as a dictionary

        `Task.as_dict()['params']`

        Returns:
            Dict with object content
        """
        return self._json

    def refresh(self):
        """Refreshes the task details."""
        self._json = self._client.get_task(self.id).as_dict()

    def cancel(self, clear_unique_id: bool = False):
        """Cancels the task"""
        self._client.cancel_task(self.id, clear_unique_id)

    def update_unique_id(self, unique_id: str):
        """Updates unique_id of a task"""
        self._client.update_task_unique_id(self.id, unique_id)

    def clear_unique_id(self):
        """Clears unique_id of a task"""
        self._client.clear_task_unique_id(self.id)
